#-----------------------------------------------------------------------------------------------------------------------------------
# Name:           Automation of input data representing marked routes in maps
# Purpose:        Generating coloured hiking routes along lines
# Attachment:     2.2
#
# Author:         Jakub Pospíšil
# University:     Palacký University Olomouc
# Purpose:        Diploma thesis
#
# Script v.:      1.5
# Created:        08.05.2024
# Licence:        CC BY 4.0
# Licence text:   Automation of input data representing marked routes in maps © 2024 by Bc. Jakub Pospíšil is licensed under CC BY 4.0 
#-----------------------------------------------------------------------------------------------------------------------------------

# Pseudocode:
# This script is designed to generate hiking and skiing routes with customizable features such as colors, width
# and offsets. It utilizes input layers containing route data and allows users to specify parameters such as
# reference scale and line width. # The script offers options to generate routes either with symbology using
# offsets or parallel lines, where features are physically moved. Additionally, it provides flexibility in naming
# output layers and choosing whether to include hiking or skiing routes in the generated data. Overall, it
# streamlines the processof creating visually appealing and informative route maps within the ArcGIS environment.

# Usage:
# This script is mainly for data from OpenStreetMap, but also works for other datasets. Input data from
# OpenStreetMap should consist of 3 attribute fields: route, colour and osmc_symbo. Those data can be downloaded
# using Relation:route. Data can be downloaded by QuickOSM plugin in QGIS using Key = route. If you choose to use
# data from other dataset, make sure it consist of at least colour attribute.

# Parameters                                                                                                            DATA TYPE       TYPE       DIRECTION   FILTER

# ap.env.workspace: Workspace environment variable for the directory of output, and in some cases also input data.      Workspace       REQUIRED   Input       No filter
# INPUT_LAYER_HIKING: Path to the input layer containing hiking trail features.                                         Feature Layer   OPTIONAL   Input       Feature type - Polyline
# INPUT_LAYER_SKIING: Path to the input layer containing skiing trail features.                                         Feature Layer   OPTIONAL   Input       Feature type - Polyline
# REFERENCE_SCALE: Reference scale affecting the resulting shift values of generated trails using parallel lines.       Long            OPTIONAL   Input       Range - Minimum - 1      - Maximum - 100000000
# LINE_WIDTH: Width of the trails line symbols.                                                                         Double          REQUIRED   Input       Range - Minimum - 0,1    - Maximum - 1000
# HIKING_COLOURED_WITH_SYMBOLOGY: Output name for hiking trails generated using offsets in symbology.                   Feature Class   OPTIONAL   Output      No filter
# HIKING_COLOURED_WITH_PARALLEL: Output name for hiking trails generated using paralle lines.                           Feature Class   OPTIONAL   Output      No filter
# SKIING_COLOURED_WITH_SYMBOLOGY: Output name for skiing trails generated using offsets in symbology.                   Feature Class   OPTIONAL   Output      No filter
# SKIING_COLOURED_WITH_PARALLEL: Output name for skiing trails generated using parallel lines.                          Feature Class   OPTIONAL   Output      No filter

import arcpy as ap

# Setting parameters that will be defined by the user
try:
    ap.env.workspace = ap.GetParameterAsText(0)
    INPUT_LAYER_HIKING = ap.GetParameterAsText(1)
    INPUT_LAYER_SKIING = ap.GetParameterAsText(2)
    REFERENCE_SCALE = ap.GetParameterAsText(3)
    LINE_WIDTH = float(ap.GetParameterAsText(4).replace(",", ".")) # Converted to float
    HIKING_COLOURED_WITH_SYMBOLOGY = ap.GetParameterAsText(5)
    HIKING_COLOURED_WITH_PARALLEL = ap.GetParameterAsText(6)
    SKIING_COLOURED_WITH_SYMBOLOGY = ap.GetParameterAsText(7)
    SKIING_COLOURED_WITH_PARALLEL = ap.GetParameterAsText(8)
except Exception as ex:
    ap.AddMessage(f"Error while obtaining input parameters: {str(ex)}")
    exit()


# Defining variables
unique_colours = {"no_colour"} # A set of colours present in the input dataset
colour_order = ["red", "blue", "green", "yellow", "black", "brown", "orange", "purple", "white", "gold", "pink", "no_colour"] # Defined colour order
map_colour_order = ["no_colour", "pink", "gold", "white", "purple", "orange", "brown", "black", "yellow", "green", "blue", "red"] # Reversed colour order
colour_rgb_values = [
    [255, 0, 0, 100],    # Red
    [0, 130, 225, 100],  # Blue
    [0, 128, 0, 100],    # Green
    [255, 255, 0, 100],  # Yellow
    [0, 0, 0, 100],      # Black
    [165, 42, 42, 100],  # Brown
    [255, 165, 0, 100],  # Orange
    [128, 0, 128, 100],  # Purple
    [255, 255, 255, 100],# White
    [255, 215, 0, 100],  # Gold
    [255, 150, 200, 100],# Pink
    [200, 200, 200, 100] # No Colour = grey
] # RGB values for colours in colour_order
colour_rgb_dictionary = dict(zip(colour_order, colour_rgb_values)) # Dictionary for colours and their RGB values
aprx = ap.mp.ArcGISProject("CURRENT") # Current project variable
aprxMap = aprx.activeMap
shift_pointer = {} # Storage of occupied shifts
colour_exists = False
osmc_symbo_exists = False
spatial_reference = False
generate_hiking_paths = False
generate_skiing_paths = False
if REFERENCE_SCALE: # Variables connected to REFERENCE SCALE
    distance = 0.00035304 * float(REFERENCE_SCALE) * LINE_WIDTH
    aprxMap.referenceScale = float(REFERENCE_SCALE) # Setting the reference scale of current map
    ap.AddMessage("Reference scale has been set to: " + REFERENCE_SCALE + ".")
    direction_hiking = "left" # Default value for generating hiking trails alongside roads
    direction_skiing = "right" # Default value for generating skiing trails alongside roads

# Temporary output feature class variables
TEMP_HIKING = "TEMP_HIKING"
TEMP_HIKING_ROADS = "TEMP_HIKING_ROADS"
TEMP_HIKING_SPLIT = "TEMP_HIKING_SPLIT"
TEMP_HIKING_COUNT = "TEMP_HIKING_COUNT"
TEMP_HIKING_IDENTIFIED = "TEMP_HIKING_IDENTIFIED"
TEMP_HIKING_DISSOLVED = "TEMP_HIKING_DISSOLVED"
TEMP_SKIING = "TEMP_SKIING"
TEMP_SKIING_ROADS = "TEMP_SKIING_ROADS"
TEMP_SKIING_COUNT = "TEMP_SKIING_COUNT"

# List of Temporary layers to be deleted at the end of the script
layers_to_delete = [
    "TEMP_HIKING",
    "TEMP_HIKING_SPLIT",
    "TEMP_HIKING_COUNT",
    "TEMP_HIKING_IDENTIFIED",
    "TEMP_HIKING_DISSOLVED",
    "TEMP_SKIING",
    "TEMP_SKIING_COUNT"
]

# Definition of functions that defines optimal EPSG code and creates parallel lines
def get_optimal_utm_zone(extent):
    ''' This function determines the optimal UTM zone and hemisphere for a given geographical extent, based on its central point, and returns a UTM zone number and detected hemisphere. '''
    # Determine the central point of the extent
    central_point_x = (extent.XMin + extent.XMax) / 2
    central_point_y = (extent.YMin + extent.YMax) / 2

    # Determine the appropriate UTM zone and hemisphere based on the central points
    hemisphere = 'N' if central_point_y >= 0 else 'S'
    utm_zone = int((central_point_x + 180) / 6) + 1

    return utm_zone, hemisphere

def create_parallel(input_line, parallel_value, parallel_direction):
    ''' This function generates a parallel polyline to a given input polyline, with an offset determined by the Shift_value parameter and in the specified direction (left or right) '''
    line_part = input_line.getPart(0)
    new_point_array = ap.Array()
    
    for point in line_part:
        distance = input_line.measureOnLine(point)
        point0 = input_line.positionAlongLine(distance - 0.01).firstPoint
        point1 = input_line.positionAlongLine(distance + 0.01).firstPoint
        difference_x = float(point1.X) - float(point0.X)
        difference_y = float(point1.Y) - float(point0.Y)
        length = (difference_x ** 2 + difference_y ** 2) ** 0.5

        horizontal_move = difference_y * float(parallel_value) / length
        vertical_move = -difference_x * float(parallel_value) / length

        if parallel_direction == "left":
            horizontal_move *= -1
            vertical_move *= -1

        new_point = ap.Point(point.X + horizontal_move, point.Y + vertical_move)
        new_point_array.add(new_point)

    parallel_line = ap.Polyline(ap.Array([new_point_array]))

    return parallel_line

# Check if user defined at least one input layer
if not INPUT_LAYER_HIKING and not INPUT_LAYER_SKIING:
    ap.AddError("Script has no input layers defined. If you want to create parallel lines, you must define at least one input layer.")
    exit()

# Check if hiking output layer is defined
if INPUT_LAYER_HIKING and not (HIKING_COLOURED_WITH_SYMBOLOGY or HIKING_COLOURED_WITH_PARALLEL):
    ap.AddWarning("Script has no output layers defined for hiking input layer.")

# Option for generating hiking routes
if INPUT_LAYER_HIKING and (HIKING_COLOURED_WITH_SYMBOLOGY or HIKING_COLOURED_WITH_PARALLEL):

    # Check if reference scale is defined if user wants to generate hiking routes using parallel lines
    if HIKING_COLOURED_WITH_PARALLEL and not REFERENCE_SCALE:
        ap.AddError("Reference scale must be set to generate hiking parallel lines. Shift distances for parallel lines will not be determined. If you want parallel lines, please set the value in the reference scale box.")
        exit()

    if HIKING_COLOURED_WITH_PARALLEL:
        if HIKING_COLOURED_WITH_PARALLEL == INPUT_LAYER_HIKING or HIKING_COLOURED_WITH_PARALLEL == INPUT_LAYER_SKIING:
            ap.AddWarning("Output hiking trails have the same name as one of input layers. Output layer name will consist of: _output")
            HIKING_COLOURED_WITH_PARALLEL = HIKING_COLOURED_WITH_PARALLEL + "_output"

    if HIKING_COLOURED_WITH_SYMBOLOGY:
        if HIKING_COLOURED_WITH_SYMBOLOGY == INPUT_LAYER_HIKING or HIKING_COLOURED_WITH_SYMBOLOGY == INPUT_LAYER_SKIING:
            ap.AddWarning("Output hiking trails have the same name as one of input layers. Output layer name will consist of: _output")
            HIKING_COLOURED_WITH_SYMBOLOGY = HIKING_COLOURED_WITH_SYMBOLOGY + "_output"

    if HIKING_COLOURED_WITH_SYMBOLOGY == HIKING_COLOURED_WITH_PARALLEL:
        ap.AddWarning("Both output hiking trails have the same name. Hiking trails using parallel lines will be named: " + HIKING_COLOURED_WITH_SYMBOLOGY.split("\\")[-1] + "_parallel.")
        HIKING_COLOURED_WITH_PARALLEL = HIKING_COLOURED_WITH_SYMBOLOGY + "_parallel"

    # Check if input layer is Polyline (not necessary now, because filter is used for input)
    #describe_hiking = ap.Describe(INPUT_LAYER_HIKING)
    #if describe_hiking.shapeType != 'Polyline':
    #    ap.AddError("Input layer for hiking trails must be a polyline feature class.")
    #    exit()

    ap.AddMessage("Hiking trails will be part of the final result.")

    # Copy feature so that original file remains untouched using project tool to make sure layer projection is in degrees
    ap.Project_management(INPUT_LAYER_HIKING, TEMP_HIKING, "4326")
    ap.AddMessage("Input feature class has been copied succesfully and projection has been set.")

    # Get the extent of the input layer
    describe_hiking = ap.Describe(TEMP_HIKING)
    extent = describe_hiking.extent
    extent_message = f"Extent of selected layer is: XMin: {extent.XMin}, YMin: {extent.YMin}, XMax: {extent.XMax}, YMax: {extent.YMax}"
    ap.AddMessage(extent_message)

    # Calculate the optimal UTM zone and use Project tool to apply it
    optimal_utm_zone, hemisphere = get_optimal_utm_zone(extent)
    # Create a spatial reference object for the optimal UTM zone and get EPSG code
    spatial_reference_hiking = ap.SpatialReference()
    spatial_reference_hiking.factoryCode = 32600 + optimal_utm_zone if hemisphere == 'N' else 32700 + optimal_utm_zone
    # Project the input layer to the optimal UTM zone with determined EPSG code
    ap.Project_management(TEMP_HIKING, TEMP_HIKING_ROADS, spatial_reference_hiking)

    # Create a Describe object for the layer TEMP_HIKING_ROADS
    describe_project = ap.Describe(TEMP_HIKING_ROADS)
    # Get the EPSG code of projected layer and use it as a new spatial reference
    epsg_code_project = describe_project.spatialReference.factoryCode
    ap.AddMessage(f"Calculated EPSG code for output feature class is: {epsg_code_project}. Spatial reference will be updated.")
    new_spatial_reference = ap.SpatialReference(epsg_code_project)
    aprxMap.spatialReference = new_spatial_reference
    spatial_reference = True

    # Create a list of fields and check if "route", "colour" and "osmc_symbo" fields exists
    fields = ap.ListFields(TEMP_HIKING_ROADS)
    for field in fields:
        if field.name == "route":
            ap.AddMessage("Route field has been succesfully found.")
            # Selecting features based on the where clause and deleting them. It is done to select only hiking and foot line features
            HIKING_PROJECT_LAYER = ap.management.MakeFeatureLayer(TEMP_HIKING_ROADS, "HIKING_PROJECT_LAYER", "route <> 'hiking' AND route <> 'foot'")
            ap.management.DeleteFeatures(HIKING_PROJECT_LAYER)
        if field.name.lower() == "colour":
            colour_exists = True
            ap.AddMessage("Colour field has been succesfully found.")
        if field.name == "osmc_symbo":
            osmc_symbo_exists = True
            ap.AddMessage("Osmc_symbo field has been succesfully found.")

    # Add a field to store colors and determine them from colour or osmc_symbo
    ap.management.AddField(TEMP_HIKING_ROADS, "colour_identified", "TEXT")
    ap.AddMessage("New field for colour values has been created.")
    # If there are no colour and osmc_symbo fields
    if not colour_exists and not osmc_symbo_exists:
        ap.AddWarning ("There are no fields colour or osmc_symbo in attributes of the input feature class. Colours could not be determined.")
        with ap.da.UpdateCursor(TEMP_HIKING_ROADS, ["colour_identified"]) as cursor:
            for row in cursor:
                row[0] = "no_colour"
    # Check colour field first and determine values
    if colour_exists:
        with ap.da.UpdateCursor(TEMP_HIKING_ROADS, ["colour", "colour_identified"]) as cursor:
            for row in cursor:
                if row[0] is None:
                    row[1] = "no_colour"
                else:    
                    if row[0].strip().lower() not in colour_order:
                        row[1] = "no_colour"
                    else:
                        # Add the color to the set
                        unique_colours.add(row[0].lower())
                        # Copy colour to the new field
                        row[1] = row[0].lower()
                cursor.updateRow(row)
    # Secondly, check osmc_symbo field determine missing values
    if osmc_symbo_exists:
        with ap.da.UpdateCursor(TEMP_HIKING_ROADS, ["osmc_symbo", "colour_identified"]) as cursor:
            for row in cursor:
                if row[0] is None:
                    if row[1] is None:
                        row[1] = "no_colour"
                elif row[0] is not None:
                    # Remove leading and trailing whitespace, split the string based on ':' and take the first part
                    colour = row[0].strip().split(':')[0].lower()
                    # Add the color to the set
                    unique_colours.add(colour)
                    # Copy colour to the new field
                    if row[1] == "no_colour" and (" " not in row[0] and ":" in row[0]):
                        row[1] = colour if colour in colour_order else "no_colour"
                    if row[1] is None:
                        row[1] = colour if colour in colour_order else "no_colour"
                cursor.updateRow(row)
    ap.AddMessage("Colours have been assigned to the new field and unnecessary fields will be deleted.")

    # Deleting unnecessary attributes
    for field in ap.ListFields(TEMP_HIKING_ROADS):
        if field.name not in ["OBJECTID", "Shape", "colour_identified", "Shape_Length"]:
            ap.management.DeleteField(TEMP_HIKING_ROADS, field.name)

    # Identity tool to determine the amount of coloured routes on the same path segment
    ap.analysis.Identity(TEMP_HIKING_ROADS, TEMP_HIKING_ROADS, TEMP_HIKING_SPLIT, "ALL")
    # Selecting features based on the where clause and deleting them. Deleting only features that have different values in colour fields
    HIKING_SPLIT_LAYER = ap.management.MakeFeatureLayer(TEMP_HIKING_SPLIT, "HIKING_SPLIT_LAYER", "colour_identified <> colour_identified_1")
    ap.management.DeleteFeatures(HIKING_SPLIT_LAYER)

    # Deleting unnecessary attributes
    for field in ap.ListFields(TEMP_HIKING_SPLIT):
        if field.name not in ["OBJECTID", "Shape", "colour_identified", "Shape_Length"]:
            ap.management.DeleteField(TEMP_HIKING_SPLIT, field.name)
    ap.AddMessage("Segments with different colour values have been identified.")

    # Count overlapping features tool in order to generate unique segment IDs
    ap.analysis.CountOverlappingFeatures(TEMP_HIKING_SPLIT, TEMP_HIKING_COUNT)
    # Identiyy tool in order to transfer segment IDs to segmnents with same line, but different colour
    ap.analysis.Identity(TEMP_HIKING_SPLIT, TEMP_HIKING_COUNT, TEMP_HIKING_IDENTIFIED, "ALL")
    ap.AddMessage("Overlapping features have been counted successfully and new segment IDs have been transfered.")
    
    # Dissolving features and optimising attributes
    ap.management.Dissolve(TEMP_HIKING_IDENTIFIED, TEMP_HIKING_DISSOLVED, ["colour_identified", "FID_TEMP_HIKING_COUNT"], [["colour_identified", "FIRST"]], "MULTI_PART")
    # Count overlapping features 2nd time in order to generate overlapping colour counts
    ap.analysis.CountOverlappingFeatures(TEMP_HIKING_DISSOLVED, TEMP_HIKING_COUNT)
    # Identity tool 2nd time in order to transfer colour counts
    ap.analysis.Identity(TEMP_HIKING_DISSOLVED, TEMP_HIKING_COUNT, TEMP_HIKING_IDENTIFIED, "ALL")
    ap.AddMessage("Features have been dissolved, overlapping features have been counted successfully and colour counts have been determined. Unnecessary fields will be deleted.")
    for field in ap.ListFields(TEMP_HIKING_IDENTIFIED):
        if field.name not in ["OBJECTID", "Shape", "FIRST_colour_identified", "COUNT_", "FID_TEMP_HIKING_COUNT", "Shape_Length"]:
            ap.management.DeleteField(TEMP_HIKING_IDENTIFIED, field.name)
    # Change field names and add new one
    for old_field, new_field in zip(["FIRST_colour_identified", "COUNT_", "FID_TEMP_HIKING_COUNT"], ["COLOUR", "COUNT", "SEGMENT_ID"]):
        ap.management.AlterField(TEMP_HIKING_IDENTIFIED, old_field, new_field, clear_field_alias = "CLEAR_ALIAS")
    ap.management.AddField(TEMP_HIKING_IDENTIFIED, "SHIFT", "SHORT")
    ap.AddMessage("Shift field has been added and other fields have been optimised.")

    # Iterating features and defining SHIFT values for single lines
    with ap.da.UpdateCursor(TEMP_HIKING_IDENTIFIED, ["SEGMENT_ID", "COUNT", "SHIFT"]) as cursor:
        for row in cursor:
            current_count = row[1]
            segment_id = row[0]

            if current_count == 1:
                row[2] = 1
                cursor.updateRow(row)
            else:
                shift_pointer[segment_id] = 1  # Adds +1 to shift_pointer for specific segment_id if there are more colours
    # Iterating ordered colours and defining SHIFT for overlapping lines            
    for ordered_colour in colour_order:
        if ordered_colour in unique_colours:
            with ap.da.UpdateCursor(TEMP_HIKING_IDENTIFIED, ["SEGMENT_ID", "COLOUR", "SHIFT"]) as cursor:
                for row in cursor:
                    segment_id = row[0]
                    colour = row[1]
                    shift = row[2]
                    if not shift and ordered_colour == colour:
                        row[2] = shift_pointer[segment_id]
                        cursor.updateRow(row)
                        shift_pointer[segment_id] += 1
    ap.AddMessage("Shift values have been determined.")

    # Add a new field to store colour order used in the map and determine values
    ap.management.AddField(TEMP_HIKING_IDENTIFIED, "colour_order", "SHORT")
    with ap.da.UpdateCursor(TEMP_HIKING_IDENTIFIED, ["colour", "colour_order"]) as cursor:
        for row in cursor:
            colour = row[0]
            if colour in map_colour_order:
                index = map_colour_order.index(colour) + 1
                row[1] = index
            cursor.updateRow(row)

    # Option for generating lines using symbology
    if HIKING_COLOURED_WITH_SYMBOLOGY:
        ap.AddMessage("Hiking trails will be generated using symbology.")
        ap.management.MultipartToSinglepart(TEMP_HIKING_IDENTIFIED, HIKING_COLOURED_WITH_SYMBOLOGY)
        # SYMBOLOGY PART
        # Add layer to the map
        lyr = aprxMap.addDataFromPath(HIKING_COLOURED_WITH_SYMBOLOGY)
        # Update renderer, RGB values and line width
        sym = lyr.symbology
        sym.updateRenderer('UniqueValueRenderer')
        sym.renderer.fields = ["colour_order", "COLOUR", "SHIFT"]
        for group in sym.renderer.groups:
            for item in group.items:
                colour_labels = item.values
                colour_label = colour_labels[0][1]
                if colour_label in colour_rgb_dictionary:
                    # Assign RGB code from colour_rgb_dictionary
                    rgb_value = colour_rgb_dictionary[colour_label]
                    item.symbol.color = {'RGB': rgb_value}
                else:
                    # Assign default grey color
                    item.symbol.color = {'RGB': [200, 200, 200, 100]}
                # Set the line offset    
                item.symbol.size = LINE_WIDTH
        # Set symbology
        lyr.symbology = sym
        # Create new CIM objects to update offsets
        sym = lyr.getDefinition('V3')
        for group in sym.renderer.groups:
            for sym_class in group.classes:
                label = sym_class.label
                shift_value = float(label.split(", ")[2])
                layer_offset = sym_class.symbol.symbol.symbolLayers[0]
                new_offset = ap.cim.CreateCIMObjectFromClassName('CIMGeometricEffectOffset', 'V3')
                new_offset.offset = LINE_WIDTH * shift_value
                # Apply new effect
                layer_offset.effects = [new_offset]
                #layer_offset.capStyle = "Butt"
        # Set symbology
        lyr.setDefinition(sym)
        # Change labels to display only colour values
        sym = lyr.symbology
        for group in sym.renderer.groups:
            for item in group.items:
                colour_labels = item.values
                colour_label = colour_labels[0][1]
                item.label = str(colour_label)
        lyr.symbology = sym

    # Option for generating lines using parallel lines
    if HIKING_COLOURED_WITH_PARALLEL and REFERENCE_SCALE:
        ap.AddMessage("Hiking trails will be generated using parallel lines.")
        ap.management.MultipartToSinglepart(TEMP_HIKING_IDENTIFIED, HIKING_COLOURED_WITH_PARALLEL)
        # Shift lines using Create_parallel function
        with ap.da.UpdateCursor(HIKING_COLOURED_WITH_PARALLEL, ("Shape@", "SHIFT")) as cursor:
            for row, shift_distance in cursor:
                generate_hiking_parallel = create_parallel(row, shift_distance * distance, direction_hiking)
                cursor.updateRow((generate_hiking_parallel, shift_distance))
        ap.AddMessage("Lines have been been shifted.")
        # Check if Repair Self Intersection tool is licensed. If so, it will be used to delete self intersections.
        if ap.CheckExtension("Foundation") == "Available":
            ap.topographic.RepairSelfIntersection(HIKING_COLOURED_WITH_PARALLEL, "DELETE")
            ap.AddMessage("Self-overlaps have been solved.")
        else:
            ap.AddMessage("Repair Self Intersection tool is not licensed. Self-overlaps have not been solved.")
        # SYMBOLOGY PART
        # Add layer to the map
        lyr = aprxMap.addDataFromPath(HIKING_COLOURED_WITH_PARALLEL)
        # Update renderer, RGB values and line width
        sym = lyr.symbology
        sym.updateRenderer('UniqueValueRenderer')
        sym.renderer.fields = ["colour_order", "COLOUR"]
        for group in sym.renderer.groups:
            for item in group.items:
                colour_labels = item.values
                colour_label = colour_labels[0][1]
                item.label = str(colour_label)
                if item.label in colour_rgb_dictionary:
                    # Assign RGB code from colour_rgb_dictionary
                    rgb_value = colour_rgb_dictionary[item.label]
                    item.symbol.color = {'RGB': rgb_value}
                    item.symbol.size = LINE_WIDTH
                else:
                    # Assign default grey color
                    item.symbol.color = {'RGB': [200, 200, 200, 100]}
                    item.symbol.size = LINE_WIDTH
        # Set symbology
        lyr.symbology = sym

    # Generate hiking paths at the end of the script
    generate_hiking_paths = True

elif not (INPUT_LAYER_HIKING and (HIKING_COLOURED_WITH_PARALLEL or HIKING_COLOURED_WITH_SYMBOLOGY)):
    ap.AddMessage("No hiking routes will be generated.")
    if not INPUT_LAYER_HIKING and (HIKING_COLOURED_WITH_PARALLEL or HIKING_COLOURED_WITH_SYMBOLOGY):
        ap.AddMessage("If you want hiking routes, you should define input layer.")
    if INPUT_LAYER_HIKING and not HIKING_COLOURED_WITH_PARALLEL and not HIKING_COLOURED_WITH_SYMBOLOGY:
        ap.AddMessage("If you want hiking routes, you should define whether you want to generate hiking routes using symbology or generated parallel lines.")

# Check if skiing output is defined
if INPUT_LAYER_SKIING and not (SKIING_COLOURED_WITH_SYMBOLOGY or SKIING_COLOURED_WITH_PARALLEL):
    ap.AddWarning("Script has no output layers defined for skiing input layer.")

# Option for generating skiing routes
if INPUT_LAYER_SKIING and (SKIING_COLOURED_WITH_SYMBOLOGY or SKIING_COLOURED_WITH_PARALLEL):

    # Check if reference scale is defined if user wants to generate skiing routes using parallel lines
    if SKIING_COLOURED_WITH_PARALLEL and not REFERENCE_SCALE:
        ap.AddError("Reference scale must be set to generate skiing parallel lines. Shift distances for parallel lines will not be determined. If you want parallel lines, please set the value in the reference scale box.")
        exit()

    if SKIING_COLOURED_WITH_SYMBOLOGY:
        if SKIING_COLOURED_WITH_SYMBOLOGY == HIKING_COLOURED_WITH_SYMBOLOGY or SKIING_COLOURED_WITH_SYMBOLOGY == HIKING_COLOURED_WITH_PARALLEL:
            ap.AddWarning("Output skiing trails have the same name as one of hiking output layers. Skiing trails output name will consist of: _skiing")
            SKIING_COLOURED_WITH_SYMBOLOGY = SKIING_COLOURED_WITH_SYMBOLOGY + "_skiing"

    if SKIING_COLOURED_WITH_SYMBOLOGY:
        if SKIING_COLOURED_WITH_SYMBOLOGY == INPUT_LAYER_HIKING or SKIING_COLOURED_WITH_SYMBOLOGY == INPUT_LAYER_SKIING:
            ap.AddWarning("Output skiing trails have same name as one of input layers. Output layer name will consist of: _output")
            SKIING_COLOURED_WITH_SYMBOLOGY = SKIING_COLOURED_WITH_SYMBOLOGY + "_output"

    if SKIING_COLOURED_WITH_PARALLEL:
        if SKIING_COLOURED_WITH_PARALLEL == HIKING_COLOURED_WITH_SYMBOLOGY or SKIING_COLOURED_WITH_PARALLEL == HIKING_COLOURED_WITH_PARALLEL:
            ap.AddWarning("Output skiing trails have the same name as one of hiking output layers. Skiing trails output name will consist of: _skiing")
            SKIING_COLOURED_WITH_PARALLEL = SKIING_COLOURED_WITH_PARALLEL + "_skiing"

    if SKIING_COLOURED_WITH_PARALLEL:
        if SKIING_COLOURED_WITH_PARALLEL == INPUT_LAYER_HIKING or SKIING_COLOURED_WITH_PARALLEL == INPUT_LAYER_SKIING:
            ap.AddWarning("Output skiing trails have same name as one of input layers. Output layer will name consist of: _output")
            SKIING_COLOURED_WITH_PARALLEL = SKIING_COLOURED_WITH_PARALLEL + "_output"

    if SKIING_COLOURED_WITH_SYMBOLOGY == SKIING_COLOURED_WITH_PARALLEL:
        ap.AddWarning("Output skiing trails have the same name. Skiing trails using parallel lines will be named: " + SKIING_COLOURED_WITH_SYMBOLOGY.split("\\")[-1] + "_parallel.")
        SKIING_COLOURED_WITH_PARALLEL = SKIING_COLOURED_WITH_SYMBOLOGY + "_parallel"

    # Check if input layer is Polyline (not necessary now, because filter is used for input)
    #describe_skiing = ap.Describe(INPUT_LAYER_SKIING)
    #if describe_skiing.shapeType != 'Polyline':
    #    ap.AddError("Input layer for skiing trails must be a polyline feature class.")
    #    exit()

    ap.AddMessage("Skiing trails will be part of final result.")

    # Copy feature so that original file remains untouched using project tool to make sure layer projection is in degrees
    ap.Project_management(INPUT_LAYER_SKIING, TEMP_SKIING, "4326")
    ap.AddMessage("Input feature class has been copied succesfully and projection has been set.")

    # Get the extent of the input layer
    describe_skiing = ap.Describe(TEMP_SKIING)
    extent = describe_skiing.extent
    extent_message = f"Extent of selected layer is: XMin: {extent.XMin}, YMin: {extent.YMin}, XMax: {extent.XMax}, YMax: {extent.YMax}"
    ap.AddMessage(extent_message)

    # Calculate the optimal UTM zone and use Project tool to apply it
    optimal_utm_zone, hemisphere = get_optimal_utm_zone(extent)
    # Create a spatial reference object for the optimal UTM zone and get EPSG code
    spatial_reference_skiing = ap.SpatialReference()
    spatial_reference_skiing.factoryCode = 32600 + optimal_utm_zone if hemisphere == 'N' else 32700 + optimal_utm_zone # UTM zone spatial reference factory code
    # Project the input layer to the optimal UTM zone with determined EPSG code
    ap.Project_management(TEMP_SKIING, TEMP_SKIING_ROADS, spatial_reference_skiing)

    # Create a Describe object for the layer TEMP_SKIING_ROADS
    describe_project = ap.Describe(TEMP_SKIING_ROADS)
    # Get the EPSG code of projected layer and use it as a new spatial reference in case it has not been done yet
    epsg_code_project = describe_project.spatialReference.factoryCode
    ap.AddMessage(f"Calculated EPSG code for output feature class is: {epsg_code_project}. Spatial reference will be updated.")
    if spatial_reference == False:
        new_spatial_reference = ap.SpatialReference(epsg_code_project)
        aprxMap.spatialReference = new_spatial_reference

    # Create a list of fields and check if "route" field exists
    fields = ap.ListFields(TEMP_SKIING_ROADS)
    for field in fields:
        if field.name == "route":
            ap.AddMessage("Route field has been succesfully found.")
            # Selecting features based on the where clause and deleting them. It is done to select only skiing and mtb line features
            SKIING_SPLIT_LAYER = ap.management.MakeFeatureLayer(TEMP_SKIING_ROADS, "SKIING_SPLIT_LAYER", "route <> 'skiing' AND route <> 'mtb'")
            ap.management.DeleteFeatures(SKIING_SPLIT_LAYER)

    # Deleting unnecessary attributes
    #for field in ap.ListFields(SKIING_PROJECT):
    #    ap.AddMessage(field.name)
    #    if "id" not in field.name.lower() and field.name not in ["Shape", "Shape_Length"]:
    #        ap.management.DeleteField(SKIING_PROJECT, field.name)
    #ap.AddMessage("Unnecessary fields have been deleted.")

    # Option for generating lines using symbology
    if SKIING_COLOURED_WITH_SYMBOLOGY:
        ap.AddMessage("Skiing trails will be generated using symbology.")
        # Count overlapping features tool used to generate unique segment IDs
        ap.analysis.CountOverlappingFeatures(TEMP_SKIING_ROADS, SKIING_COLOURED_WITH_SYMBOLOGY)
        # SYMBOLOGY PART
        # Add layer to the map
        lyr = aprxMap.addDataFromPath(SKIING_COLOURED_WITH_SYMBOLOGY)
        # Update renderer, RGB value line width and create new CIM object to update offset
        sym = lyr.symbology
        sym.updateRenderer('SimpleRenderer')
        lyr.symbology = sym
        # Turning off labels
        if lyr.supports("SHOWLABELS"):
            # List label classes
            label_classes = lyr.listLabelClasses()
            # Delete expression so that there is no label
            for label_class in label_classes:
                label_class.expression = ""
        # Update RGB value, line width and create new CIM objects to update offset and apply dash effect
        symbology = lyr.getDefinition('V3')
        symbol = symbology.renderer.symbol.symbol.symbolLayers[0]
        symbol.color = [0, 50, 200]
        symbol.width = LINE_WIDTH
        #symbol.capStyle = "Butt"
        new_offset = ap.cim.CreateCIMObjectFromClassName('CIMGeometricEffectOffset', 'V3')
        new_offset.offset = (-1 * LINE_WIDTH)
        # Apply new effect
        symbol.effects = [new_offset]
        # Set symbology
        lyr.setDefinition(symbology)

    # Option for generating lines using parallel lines
    if SKIING_COLOURED_WITH_PARALLEL and REFERENCE_SCALE:
        ap.AddMessage("Skiing trails will be generated using parallel lines.")
        # Count overlapping features in order to generate unique segment IDs
        ap.analysis.CountOverlappingFeatures(TEMP_SKIING_ROADS, TEMP_SKIING_COUNT)
        # Multipart to singlepart tool for CopyParallel function to work properly
        ap.management.MultipartToSinglepart(TEMP_SKIING_COUNT, SKIING_COLOURED_WITH_PARALLEL)
        # Creating parallel lines for SKIING routes.
        ap.AddField_management(SKIING_COLOURED_WITH_PARALLEL, "SHIFT_VALUE", "DOUBLE")
        ap.AddMessage("Overlapping features have been counted successfully and new field for shift values has been created.")
        # Shift lines using Create_parallel function
        with ap.da.UpdateCursor(SKIING_COLOURED_WITH_PARALLEL, ("Shape@", "SHIFT_VALUE")) as cursor:
            for row, shift_distance in cursor:
                generate_skiing_parallel = create_parallel(row, distance, direction_skiing)
                cursor.updateRow((generate_skiing_parallel, distance))
        ap.AddMessage("Lines have been been shifted.")
        # Check if Repair Self Intersection tool is licensed. If so, it will be used to delete self intersections.
        if ap.CheckExtension("Foundation") == "Available":
            ap.topographic.RepairSelfIntersection(SKIING_COLOURED_WITH_PARALLEL, "DELETE")
            ap.AddMessage("Self-overlaps have been solved.")
        else:
            ap.AddMessage("Repair Self Intersection tool is not licensed. Self-overlaps have not been solved.")
        # SYMBOLOGY PART
        # Add layer to the map
        lyr = aprxMap.addDataFromPath(SKIING_COLOURED_WITH_PARALLEL)
        # Update RGB value, line width and create new CIM object to apply dash effect
        symbology = lyr.getDefinition('V3')
        symbol = symbology.renderer.symbol.symbol.symbolLayers[0]
        symbol.color = [0, 50, 200]
        symbol.width = LINE_WIDTH
        #symbol.capStyle = "Butt"        
        # Set symbology
        lyr.setDefinition(symbology)
    
    # Generate skiing paths at the end of the script
    generate_skiing_paths = True

elif not (INPUT_LAYER_SKIING and (SKIING_COLOURED_WITH_PARALLEL or SKIING_COLOURED_WITH_SYMBOLOGY)):
    ap.AddMessage("No skiing routes will be generated.")
    if not INPUT_LAYER_SKIING and (SKIING_COLOURED_WITH_PARALLEL or SKIING_COLOURED_WITH_SYMBOLOGY):
        ap.AddMessage("If you want skiing routes, you should define input layer.")
    if INPUT_LAYER_SKIING and not SKIING_COLOURED_WITH_PARALLEL and not SKIING_COLOURED_WITH_SYMBOLOGY:
        ap.AddMessage("If you want skiing routes, you should define whether you want to generate hiking routes using symbology or generated parallel lines.")

# Add paths to the map
if generate_hiking_paths and not generate_skiing_paths:
    # Add hiking paths to the map
    roads_hiking_path = ap.GetParameterAsText(0) + "\\" + TEMP_HIKING_ROADS
    roads_hiking = aprxMap.addDataFromPath(roads_hiking_path)
    # Update symbology of hiking paths
    sym = roads_hiking.getDefinition('V3')
    symbol = sym.renderer.symbol.symbol.symbolLayers[0]
    symbol.color = [153, 153, 153]
    symbol.width = LINE_WIDTH
    roads_hiking.setDefinition(sym)
    ap.AddMessage("Hiking paths feature class has been added to the map.")

if not generate_hiking_paths and generate_skiing_paths:
    # Add skiing paths to the map    
    roads_skiing_path = ap.GetParameterAsText(0) + "\\" + TEMP_SKIING_ROADS
    roads_bicyle = aprxMap.addDataFromPath(roads_skiing_path)
    #Update symbology of skiing paths
    sym = roads_bicyle.getDefinition('V3')
    symbol = sym.renderer.symbol.symbol.symbolLayers[0]
    symbol.color = [153, 153, 153]
    symbol.width = LINE_WIDTH
    roads_bicyle.setDefinition(sym)
    ap.AddMessage("Skiing paths feature class has been added to the map.")

if generate_hiking_paths and generate_skiing_paths:
    ap.management.Merge([TEMP_HIKING_ROADS, TEMP_SKIING_ROADS], "HIKING_SKIING_PATHS")
    # Add hiking paths to the map
    roads_hiking_skiing_path = ap.GetParameterAsText(0) + "\\" + "HIKING_SKIING_PATHS"
    roads_hiking_skiing = aprxMap.addDataFromPath(roads_hiking_skiing_path)
    # Update symbology of hiking paths
    sym = roads_hiking_skiing.getDefinition('V3')
    symbol = sym.renderer.symbol.symbol.symbolLayers[0]
    symbol.color = [153, 153, 153]
    symbol.width = LINE_WIDTH
    roads_hiking_skiing.setDefinition(sym)
    ap.AddMessage("Hiking paths were merged with skiing paths and resulting feature class has been added to the map.")

# Delete temporary layers
for layer_name in layers_to_delete:
    ap.management.Delete(layer_name, "FeatureClass")
ap.AddMessage("Temporary layers have been deleted.")

# Save changes in project
#aprx.save()
ap.AddMessage("Script has run successfully.")