#-------------------------------------------------------------------------------
# Name:         Snap OpenStreetMap line features
# Purpose:      Snapping line features from different data sources
# Attachment:   2.4
#
# Author:       Jakub Pospíšil
# University:   Palacký University Olomouc
# Purpose:      Diploma thesis
#
# Script v.:    1.5
# Created:      08.05.2024
# Licence:      CC BY 4.0
# Licence text: Snap OpenStreetMap line features © 2024 by Bc. Jakub Pospíšil is licensed under CC BY 4.0 
#-------------------------------------------------------------------------------

# Pseudocode:
# This arcpy Python script processes spatial tourism data. First, it defines user parameters such as the
# workspace and input/output feature classes. Then, it sets up temporary variables for storing intermediate
# results. A function called snap_points_to_nearest is defined to snap points to the nearest road
# vertexes within a specified distance threshold. The script converts multipart features into single
# parts and generates points along and road lines for snapping. It snaps points to nearby road
# vertexes, converts them back into line features, and transfers relevant attributes. Dissolution of line
# features simplifies the dataset, followed by field name adjustments for clarity. Temporary layers are deleted
# to clean up the workspace, and the script concludes by printing success messages.

# Usage:
# This script is designed to snap line features from OpenStreetMap to line features from other data source. It is
# also designed to prepare output data for second tool generating hiking and bicycle/ski routes. OpenStreetMap
# input data should consist of 4 attribute fields: osm_id, route, colour and osmc_symbo. Those data can be
# downloaded using Relation:route. Data can be downloaded by QuickOSM plugin in QGIS using Key = route. If you
# choose to use input data from other dataset instead of OpenStreetMap, keep in mind that features in output
# layer might not be dissolved due to possible missing attribute fields.

# Parameters                                                                                                           DATA TYPE       TYPE       DIRECTION   FILTER                                      DEFAULT

# ap.env.workspace: Workspace environment variable for the directory of output, and in some cases also input data      Workspace       REQUIRED   Input       No filter
# INPUT_TARGET_FEATURES: Path to the input layer containing paths, trails and roads features. (default line features)  Feature Layer   REQUIRED   Input       Feature type - Polyline
# INPUT_LINES_TO_BE_SNAPPED: Path to the input layer containing marked trails. (line features to be snapped)           Feature Layer   REQUIRED   Input       Feature type - Polyline
# OUTPUT_LINES_TO_BE_SNAPPED: Path of the output layer of snapped line features with colour values                     Feature Class   REQUIRED   Output      No filter
# THRESHOLD: The maximum distance threshold for snapping marked trails to paths, trails and roads features             Long            OPTIONAL   Input       Range - Minimum - 1 - Maximum - 100000000   50

import arcpy as ap

# Setting parameters that can be defined by user
ap.env.workspace = ap.GetParameterAsText(0)
INPUT_TARGET_FEATURES = ap.GetParameterAsText(1)
INPUT_LINES_TO_BE_SNAPPED = ap.GetParameterAsText(2)
OUTPUT_LINES_TO_BE_SNAPPED = ap.GetParameterAsText(3)
THRESHOLD = float(ap.GetParameterAsText(4))

# Temporary output feature class variables
TEMP_LINES_TO_BE_SNAPPED_SINGLE = "TEMP_LINES_TO_BE_SNAPPED_SINGLE"
TEMP_LINES_TO_BE_SNAPPED_LINE = "TEMP_LINES_TO_BE_SNAPPED_LINE"
TEMP_TARGET_FEATURES_POINTS = "TEMP_TARGET_FEATURES_POINTS"
TEMP_OUTPUT_POINT = "TEMP_OUTPUT_POINT"

# List of Temporary layers to be deleted at the end of the script
layers_to_delete = [
    "TEMP_LINES_TO_BE_SNAPPED_SINGLE",
    "TEMP_LINES_TO_BE_SNAPPED_LINE",
    "TEMP_TARGET_FEATURES_POINTS",
    "TEMP_OUTPUT_POINT"
]

# Definition of function
def snap_points_to_nearest(points, target_features, max_distance):
    ''' This function calculates the distance between points and target_features in order to snap them. '''
    search_radius = "{} Meters".format(max_distance)
    ap.analysis.Near(points, target_features, "", "LOCATION", "NO_ANGLE", "GEODESIC")

    fields = ["NEAR_X", "NEAR_Y", "NEAR_DIST", "SHAPE@"]
    with ap.da.UpdateCursor(points, fields) as cursor:
        for row in cursor:
            near_x, near_y, near_dist, shape = row

            # If the distance is less than max_distance, perform snap
            if near_dist <= max_distance:
                new_x, new_y = near_x, near_y
            else:
                # If the distance is greater than max_distance, preserve the original position of the point
                new_x, new_y = shape.firstPoint.X, shape.firstPoint.Y

            # Create a new point with the snapped values
            new_point = ap.Point(new_x, new_y)

            # Update the shape of the point
            row[3] = new_point

            # Update the row data
            cursor.updateRow(row)

# Check if input layer are Polyline (not necessary now, because filter is used for input)
# List of input layers
#input_layers = [INPUT_TARGET_FEATURES, INPUT_LINES_TO_BE_SNAPPED]
# Loop through all input layers
#for layer in input_layers:
    # Get geometry information of the layer
    #describe_layer = ap.Describe(layer)
    # Check geometry type
    #if describe_layer.shapeType != 'Polyline':
        #ap.AddError(f"The input layer {layer} must be of type polyline.")
        #exit()

if INPUT_TARGET_FEATURES == INPUT_LINES_TO_BE_SNAPPED:
    ap.AddError("Input layers must be diffetent. Please use two different datasets.")
    exit()

if INPUT_TARGET_FEATURES == OUTPUT_LINES_TO_BE_SNAPPED or INPUT_LINES_TO_BE_SNAPPED == OUTPUT_LINES_TO_BE_SNAPPED:
    ap.AddWarning("Output layer has same name as one of input layers. Output layer will consist of: _output")
    OUTPUT_LINES_TO_BE_SNAPPED = OUTPUT_LINES_TO_BE_SNAPPED + "_output"

# Multipart to singlepart tool for better segmentation
ap.management.MultipartToSinglepart(INPUT_LINES_TO_BE_SNAPPED, TEMP_LINES_TO_BE_SNAPPED_SINGLE)
ap.AddMessage("Features have been converted using Multipart to Singlepart tool.")

# Generate points both line features 
ap.management.GeneratePointsAlongLines(TEMP_LINES_TO_BE_SNAPPED_SINGLE, TEMP_OUTPUT_POINT, 'DISTANCE', Distance='3 meters')
ap.management.GeneratePointsAlongLines(INPUT_TARGET_FEATURES, TEMP_TARGET_FEATURES_POINTS, 'DISTANCE', Distance='3 meters')
ap.AddMessage("Points along both input layers have been created.")

# Call snap function and snap points
if not THRESHOLD:
    THRESHOLD = 50
snap_points_to_nearest(TEMP_OUTPUT_POINT, TEMP_TARGET_FEATURES_POINTS, THRESHOLD)
ap.AddMessage("Points have been succesfully snapped.")

# Convert output tourist points back to line features and transfres attributes
ap.management.PointsToLine(TEMP_OUTPUT_POINT, OUTPUT_LINES_TO_BE_SNAPPED, "ORIG_FID")
ap.JoinField_management(OUTPUT_LINES_TO_BE_SNAPPED, "ORIG_FID", TEMP_OUTPUT_POINT, "ORIG_FID")
ap.AddMessage("Points features have been converted back to line features and attribute fields have been transfered.")

# Delete temporary layers
for layer_name in layers_to_delete:
    ap.management.Delete(layer_name, "FeatureClass")
ap.AddMessage("Temporary layers have been deleted.")

# End of the script
ap.AddMessage("Script has run successfully.")
